@@ -45,6 +45,11 @@ DEFAULT_USER_AGENT = 'keystoneauth1/%s %s %s/%s' % (
     keystoneauth1.__version__, requests.utils.default_user_agent(),
     platform.python_implementation(), platform.python_version())
 
+# NOTE(jamielennox): Clients will likely want to print more than json. Please
+# propose a patch if you have a content type you think is reasonable to print
+# here and we'll add it to the list as required.
+_LOG_CONTENT_TYPES = set(['application/json'])
+
 _logger = utils.get_logger(__name__)
 
 
@@ -285,7 +290,7 @@ class Session(object):
     def _process_header(header):
         """Redact the secure headers to be logged."""
         secure_headers = ('authorization', 'x-auth-token',
-                          'x-subject-token',)
+                          'x-subject-token', 'x-service-token')
         if header[0].lower() in secure_headers:
             token_hasher = hashlib.sha1()
             token_hasher.update(header[1].encode('utf-8'))
@@ -326,7 +331,7 @@ class Session(object):
             string_parts.append(url)
 
         if headers:
-            for header in six.iteritems(headers):
+            for header in headers.items():
                 string_parts.append('-H "%s: %s"'
                                     % self._process_header(header))
         if json:
@@ -354,7 +359,19 @@ class Session(object):
             if not headers:
                 headers = response.headers
             if not text:
-                text = self._remove_service_catalog(response.text)
+                # NOTE(samueldmq): If the response does not provide enough info
+                # about the content type to decide whether it is useful and
+                # safe to log it or not, just do not log the body. Trying to
+                # read the response body anyways may result on reading a long
+                # stream of bytes and getting an unexpected MemoryError. See
+                # bug 1616105 for further details.
+                content_type = response.headers.get('content-type', None)
+                if content_type in _LOG_CONTENT_TYPES:
+                    text = self._remove_service_catalog(response.text)
+                else:
+                    text = ('Omitted, Content-Type is set to %s. Only '
+                            '%s responses have their bodies logged.')
+                    text = text % (content_type, ', '.join(_LOG_CONTENT_TYPES))
         if json:
             text = self._json.encode(json)
 
@@ -363,7 +380,7 @@ class Session(object):
         if status_code:
             string_parts.append('[%s]' % status_code)
         if headers:
-            for header in six.iteritems(headers):
+            for header in headers.items():
                 string_parts.append('%s: %s' % self._process_header(header))
         if text:
             string_parts.append('\nRESP BODY: %s\n' % text)
