@@ -13,6 +13,7 @@
 import itertools
 import json
 import logging
+import sys
 import uuid
 
 import mock
@@ -133,6 +134,24 @@ class SessionTests(utils.TestCase):
         self.assertTrue(resp.ok)
         self.assertRequestHeaderEqual('User-Agent', 'overrides-agent')
 
+        # If sys.argv is an empty list, then doesn't fail.
+        with mock.patch.object(sys, 'argv', []):
+            session = client_session.Session()
+            resp = session.get(self.TEST_URL)
+            self.assertTrue(resp.ok)
+            self.assertRequestHeaderEqual(
+                'User-Agent',
+                client_session.DEFAULT_USER_AGENT)
+
+        # If sys.argv[0] is an empty string, then doesn't fail.
+        with mock.patch.object(sys, 'argv', ['']):
+            session = client_session.Session()
+            resp = session.get(self.TEST_URL)
+            self.assertTrue(resp.ok)
+            self.assertRequestHeaderEqual(
+                'User-Agent',
+                client_session.DEFAULT_USER_AGENT)
+
     def test_http_session_opts(self):
         session = client_session.Session(cert='cert.pem', timeout=5,
                                          verify='certs')
@@ -169,12 +188,14 @@ class SessionTests(utils.TestCase):
         in order to redact secure headers while debug is true.
         """
         session = client_session.Session(verify=False)
-        headers = {'HEADERA': 'HEADERVALB'}
+        headers = {'HEADERA': 'HEADERVALB',
+                   'Content-Type': 'application/json'}
         security_headers = {'Authorization': uuid.uuid4().hex,
                             'X-Auth-Token': uuid.uuid4().hex,
-                            'X-Subject-Token': uuid.uuid4().hex, }
-        body = 'BODYRESPONSE'
-        data = 'BODYDATA'
+                            'X-Subject-Token': uuid.uuid4().hex,
+                            'X-Service-Token': uuid.uuid4().hex}
+        body = '{"a": "b"}'
+        data = '{"c": "d"}'
         all_headers = dict(
             itertools.chain(headers.items(), security_headers.items()))
         self.stub_url('POST', text=body, headers=all_headers)
@@ -187,13 +208,13 @@ class SessionTests(utils.TestCase):
         self.assertIn(body, self.logger.output)
         self.assertIn("'%s'" % data, self.logger.output)
 
-        for k, v in six.iteritems(headers):
+        for k, v in headers.items():
             self.assertIn(k, self.logger.output)
             self.assertIn(v, self.logger.output)
 
         # Assert that response headers contains actual values and
         # only debug logs has been masked
-        for k, v in six.iteritems(security_headers):
+        for k, v in security_headers.items():
             self.assertIn('%s: {SHA1}' % k, self.logger.output)
             self.assertEqual(v, resp.headers[k])
             self.assertNotIn(v, self.logger.output)
@@ -201,13 +222,48 @@ class SessionTests(utils.TestCase):
     def test_logs_failed_output(self):
         """Test that output is logged even for failed requests."""
         session = client_session.Session()
-        body = uuid.uuid4().hex
+        body = {uuid.uuid4().hex: uuid.uuid4().hex}
 
-        self.stub_url('GET', text=body, status_code=400)
+        self.stub_url('GET', json=body, status_code=400,
+                      headers={'Content-Type': 'application/json'})
         resp = session.get(self.TEST_URL, raise_exc=False)
 
         self.assertEqual(resp.status_code, 400)
+        self.assertIn(list(body.keys())[0], self.logger.output)
+        self.assertIn(list(body.values())[0], self.logger.output)
+
+    def test_logging_body_only_for_specified_content_types(self):
+        """Verify response body is only logged in specific content types.
+
+        Response bodies are logged only when the response's Content-Type header
+        is set to application/json. This prevents us to get an unexpected
+        MemoryError when reading arbitrary responses, such as streams.
+        """
+        OMITTED_BODY = ('Omitted, Content-Type is set to %s. Only '
+                        'application/json responses have their bodies logged.')
+        session = client_session.Session(verify=False)
+
+        # Content-Type is not set
+        body = json.dumps({'token': {'id': '...'}})
+        self.stub_url('POST', text=body)
+        session.post(self.TEST_URL)
+        self.assertNotIn(body, self.logger.output)
+        self.assertIn(OMITTED_BODY % None, self.logger.output)
+
+        # Content-Type is set to text/xml
+        body = '<token><id>...</id></token>'
+        self.stub_url('POST', text=body, headers={'Content-Type': 'text/xml'})
+        session.post(self.TEST_URL)
+        self.assertNotIn(body, self.logger.output)
+        self.assertIn(OMITTED_BODY % 'text/xml', self.logger.output)
+
+        # Content-Type is set to application/json
+        body = json.dumps({'token': {'id': '...'}})
+        self.stub_url('POST', text=body,
+                      headers={'Content-Type': 'application/json'})
+        session.post(self.TEST_URL)
         self.assertIn(body, self.logger.output)
+        self.assertNotIn(OMITTED_BODY % 'application/json', self.logger.output)
 
     def test_logging_cacerts(self):
         path_to_certs = '/path/to/certs'
@@ -265,6 +321,25 @@ class SessionTests(utils.TestCase):
                                session.get,
                                self.TEST_URL)
 
+    def test_json_content_type(self):
+        session = client_session.Session()
+        self.stub_url('POST', text='response')
+        resp = session.post(
+            self.TEST_URL,
+            json=[{'op': 'replace',
+                   'path': '/name',
+                   'value': 'new_name'}],
+            headers={'Content-Type': 'application/json-patch+json'})
+
+        self.assertEqual('POST', self.requests_mock.last_request.method)
+        self.assertEqual(resp.text, 'response')
+        self.assertTrue(resp.ok)
+        self.assertRequestBodyIs(
+            json=[{'op': 'replace',
+                   'path': '/name',
+                   'value': 'new_name'}])
+        self.assertContentTypeIs('application/json-patch+json')
+
 
 class RedirectTests(utils.TestCase):
 
@@ -343,6 +418,12 @@ class RedirectTests(utils.TestCase):
             self.assertEqual(r.url, s.url)
             self.assertEqual(r.status_code, s.status_code)
 
+    def test_permanent_redirect_308(self):
+        session = client_session.Session()
+        self.setup_redirects(status_code=308)
+        resp = session.get(self.REDIRECT_CHAIN[-2])
+        self.assertResponse(resp)
+
 
 class AuthPlugin(plugin.BaseAuthPlugin):
     """Very simple debug authentication plugin.
@@ -390,6 +471,7 @@ class AuthPlugin(plugin.BaseAuthPlugin):
 class CalledAuthPlugin(plugin.BaseAuthPlugin):
 
     ENDPOINT = 'http://fakeendpoint/'
+    TOKEN = utils.TestCase.TEST_TOKEN
     USER_ID = uuid.uuid4().hex
     PROJECT_ID = uuid.uuid4().hex
 
@@ -404,7 +486,7 @@ class CalledAuthPlugin(plugin.BaseAuthPlugin):
 
     def get_token(self, session):
         self.get_token_called = True
-        return utils.TestCase.TEST_TOKEN
+        return self.TOKEN
 
     def get_endpoint(self, session, **kwargs):
         self.get_endpoint_called = True
@@ -711,22 +793,24 @@ class SessionAuthTests(utils.TestCase):
 
         auth = AuthPlugin()
         sess = client_session.Session(auth=auth)
-        response = uuid.uuid4().hex
+        response = {uuid.uuid4().hex: uuid.uuid4().hex}
 
         self.stub_url('GET',
-                      text=response,
-                      headers={'Content-Type': 'text/html'})
+                      json=response,
+                      headers={'Content-Type': 'application/json'})
 
         resp = sess.get(self.TEST_URL, logger=logger)
 
-        self.assertEqual(response, resp.text)
+        self.assertEqual(response, resp.json())
         output = io.getvalue()
 
         self.assertIn(self.TEST_URL, output)
-        self.assertIn(response, output)
+        self.assertIn(list(response.keys())[0], output)
+        self.assertIn(list(response.values())[0], output)
 
         self.assertNotIn(self.TEST_URL, self.logger.output)
-        self.assertNotIn(response, self.logger.output)
+        self.assertNotIn(list(response.keys())[0], self.logger.output)
+        self.assertNotIn(list(response.values())[0], self.logger.output)
 
 
 class AdapterTest(utils.TestCase):
@@ -743,11 +827,9 @@ class AdapterTest(utils.TestCase):
 
     TEST_URL = CalledAuthPlugin.ENDPOINT
 
-    def _create_loaded_adapter(self):
-        auth = CalledAuthPlugin()
-        sess = client_session.Session()
-        return adapter.Adapter(sess,
-                               auth=auth,
+    def _create_loaded_adapter(self, sess=None, auth=None):
+        return adapter.Adapter(sess or client_session.Session(),
+                               auth=auth or CalledAuthPlugin(),
                                service_type=self.SERVICE_TYPE,
                                service_name=self.SERVICE_NAME,
                                interface=self.INTERFACE,
@@ -907,21 +989,23 @@ class AdapterTest(utils.TestCase):
         sess = client_session.Session(auth=auth)
         adpt = adapter.Adapter(sess, auth=auth, logger=logger)
 
-        response = uuid.uuid4().hex
+        response = {uuid.uuid4().hex: uuid.uuid4().hex}
 
-        self.stub_url('GET', text=response,
-                      headers={'Content-Type': 'text/html'})
+        self.stub_url('GET', json=response,
+                      headers={'Content-Type': 'application/json'})
 
         resp = adpt.get(self.TEST_URL, logger=logger)
 
-        self.assertEqual(response, resp.text)
+        self.assertEqual(response, resp.json())
         output = io.getvalue()
 
         self.assertIn(self.TEST_URL, output)
-        self.assertIn(response, output)
+        self.assertIn(list(response.keys())[0], output)
+        self.assertIn(list(response.values())[0], output)
 
         self.assertNotIn(self.TEST_URL, self.logger.output)
-        self.assertNotIn(response, self.logger.output)
+        self.assertNotIn(list(response.keys())[0], self.logger.output)
+        self.assertNotIn(list(response.values())[0], self.logger.output)
 
     def test_unknown_connection_error(self):
         self.stub_url('GET', exc=requests.exceptions.RequestException)
@@ -980,6 +1064,161 @@ class AdapterTest(utils.TestCase):
         self.assertEqual(request_val,
                          self.requests_mock.last_request.headers[header])
 
+    def test_adapter_user_agent_session_adapter(self):
+        sess = client_session.Session(app_name='ksatest', app_version='1.2.3')
+        adap = adapter.Adapter(client_name='testclient',
+                               client_version='4.5.6',
+                               session=sess)
+
+        url = 'http://keystone.test.com'
+        self.requests_mock.get(url)
+
+        adap.get(url)
+
+        agent = 'ksatest/1.2.3 testclient/4.5.6'
+        self.assertEqual(agent + ' ' + client_session.DEFAULT_USER_AGENT,
+                         self.requests_mock.last_request.headers['User-Agent'])
+
+    def test_adapter_user_agent_session_version_on_adapter(self):
+
+        class TestAdapter(adapter.Adapter):
+
+            client_name = 'testclient'
+            client_version = '4.5.6'
+
+        sess = client_session.Session(app_name='ksatest', app_version='1.2.3')
+        adap = TestAdapter(session=sess)
+
+        url = 'http://keystone.test.com'
+        self.requests_mock.get(url)
+
+        adap.get(url)
+
+        agent = 'ksatest/1.2.3 testclient/4.5.6'
+        self.assertEqual(agent + ' ' + client_session.DEFAULT_USER_AGENT,
+                         self.requests_mock.last_request.headers['User-Agent'])
+
+    def test_adapter_user_agent_session_adapter_no_app_version(self):
+        sess = client_session.Session(app_name='ksatest')
+        adap = adapter.Adapter(client_name='testclient',
+                               client_version='4.5.6',
+                               session=sess)
+
+        url = 'http://keystone.test.com'
+        self.requests_mock.get(url)
+
+        adap.get(url)
+
+        agent = 'ksatest testclient/4.5.6'
+        self.assertEqual(agent + ' ' + client_session.DEFAULT_USER_AGENT,
+                         self.requests_mock.last_request.headers['User-Agent'])
+
+    def test_adapter_user_agent_session_adapter_no_client_version(self):
+        sess = client_session.Session(app_name='ksatest', app_version='1.2.3')
+        adap = adapter.Adapter(client_name='testclient', session=sess)
+
+        url = 'http://keystone.test.com'
+        self.requests_mock.get(url)
+
+        adap.get(url)
+
+        agent = 'ksatest/1.2.3 testclient'
+        self.assertEqual(agent + ' ' + client_session.DEFAULT_USER_AGENT,
+                         self.requests_mock.last_request.headers['User-Agent'])
+
+    def test_adapter_user_agent_session_adapter_additional(self):
+        sess = client_session.Session(app_name='ksatest',
+                                      app_version='1.2.3',
+                                      additional_user_agent=[('one', '1.1.1'),
+                                                             ('two', '2.2.2')])
+        adap = adapter.Adapter(client_name='testclient',
+                               client_version='4.5.6',
+                               session=sess)
+
+        url = 'http://keystone.test.com'
+        self.requests_mock.get(url)
+
+        adap.get(url)
+
+        agent = 'ksatest/1.2.3 one/1.1.1 two/2.2.2 testclient/4.5.6'
+        self.assertEqual(agent + ' ' + client_session.DEFAULT_USER_AGENT,
+                         self.requests_mock.last_request.headers['User-Agent'])
+
+    def test_adapter_user_agent_session(self):
+        sess = client_session.Session(app_name='ksatest', app_version='1.2.3')
+        adap = adapter.Adapter(session=sess)
+
+        url = 'http://keystone.test.com'
+        self.requests_mock.get(url)
+
+        adap.get(url)
+
+        agent = 'ksatest/1.2.3'
+        self.assertEqual(agent + ' ' + client_session.DEFAULT_USER_AGENT,
+                         self.requests_mock.last_request.headers['User-Agent'])
+
+    def test_adapter_user_agent_adapter(self):
+        sess = client_session.Session()
+        adap = adapter.Adapter(client_name='testclient',
+                               client_version='4.5.6',
+                               session=sess)
+
+        url = 'http://keystone.test.com'
+        self.requests_mock.get(url)
+
+        adap.get(url)
+
+        agent = 'testclient/4.5.6'
+        self.assertEqual(agent + ' ' + client_session.DEFAULT_USER_AGENT,
+                         self.requests_mock.last_request.headers['User-Agent'])
+
+    def test_adapter_user_agent_session_override(self):
+        sess = client_session.Session(app_name='ksatest',
+                                      app_version='1.2.3',
+                                      additional_user_agent=[('one', '1.1.1'),
+                                                             ('two', '2.2.2')])
+        adap = adapter.Adapter(client_name='testclient',
+                               client_version='4.5.6',
+                               session=sess)
+
+        url = 'http://keystone.test.com'
+        self.requests_mock.get(url)
+
+        override_user_agent = '%s/%s' % (uuid.uuid4().hex, uuid.uuid4().hex)
+        adap.get(url, user_agent=override_user_agent)
+
+        self.assertEqual(override_user_agent,
+                         self.requests_mock.last_request.headers['User-Agent'])
+
+    def test_nested_adapters(self):
+        text = uuid.uuid4().hex
+        token = uuid.uuid4().hex
+        url = 'http://keystone.example.com/path'
+
+        sess = client_session.Session()
+        auth = CalledAuthPlugin()
+        auth.ENDPOINT = url
+        auth.TOKEN = token
+
+        adap1 = adapter.Adapter(session=sess,
+                                interface='public')
+        adap2 = adapter.Adapter(session=adap1,
+                                service_type='identity',
+                                auth=auth)
+
+        self.requests_mock.get(url + '/test', text=text)
+
+        resp = adap2.get('/test')
+
+        self.assertEqual(text, resp.text)
+        self.assertTrue(auth.get_endpoint_called)
+
+        self.assertEqual('public', auth.endpoint_arguments['interface'])
+        self.assertEqual('identity', auth.endpoint_arguments['service_type'])
+
+        last_token = self.requests_mock.last_request.headers['X-Auth-Token']
+        self.assertEqual(token, last_token)
+
 
 class TCPKeepAliveAdapterTest(utils.TestCase):
 
@@ -1035,6 +1274,7 @@ class TCPKeepAliveAdapterTest(utils.TestCase):
 
     def test_init_poolmanager_with_tcp_keepcnt(self):
         self.patch(client_session, 'REQUESTS_VERSION', (2, 4, 1))
+        self.patch(client_session.utils, 'is_windows_linux_subsystem', False)
         socket = self.patch_socket_with_options(
             ['IPPROTO_TCP', 'TCP_NODELAY', 'SOL_SOCKET', 'SO_KEEPALIVE',
              'TCP_KEEPCNT'])
@@ -1050,6 +1290,23 @@ class TCPKeepAliveAdapterTest(utils.TestCase):
                 (socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1),
                 (socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 4)])
 
+    def test_init_poolmanager_with_tcp_keepcnt_on_windows(self):
+        self.patch(client_session, 'REQUESTS_VERSION', (2, 4, 1))
+        self.patch(client_session.utils, 'is_windows_linux_subsystem', True)
+        socket = self.patch_socket_with_options(
+            ['IPPROTO_TCP', 'TCP_NODELAY', 'SOL_SOCKET', 'SO_KEEPALIVE',
+             'TCP_KEEPCNT'])
+        given_adapter = client_session.TCPKeepAliveAdapter()
+
+        # when pool manager is initialized
+        given_adapter.init_poolmanager(1, 2, 3)
+
+        # then socket_options are given
+        self.init_poolmanager.assert_called_once_with(
+            1, 2, 3, socket_options=[
+                (socket.IPPROTO_TCP, socket.TCP_NODELAY, 1),
+                (socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)])
+
     def test_init_poolmanager_with_tcp_keepintvl(self):
         self.patch(client_session, 'REQUESTS_VERSION', (2, 4, 1))
         socket = self.patch_socket_with_options(
