@@ -18,7 +18,6 @@ import uuid
 import fixtures
 import requests
 from requests_mock.contrib import fixture
-import six
 from six.moves.urllib import parse as urlparse
 import testtools
 
@@ -98,7 +97,7 @@ class TestCase(testtools.TestCase):
         parts = urlparse.urlparse(self.requests_mock.last_request.url)
         qs = urlparse.parse_qs(parts.query, keep_blank_values=True)
 
-        for k, v in six.iteritems(kwargs):
+        for k, v in kwargs.items():
             self.assertIn(k, qs)
             self.assertIn(v, qs[k])
 
@@ -135,6 +134,12 @@ class TestResponse(requests.Response):
         """Define equiality behavior of request and response."""
         return self.__dict__ == other.__dict__
 
+    # NOTE: This function is only needed by Python 2. If we get to point where
+    # we don't support Python 2 anymore, this function should be removed.
+    def __ne__(self, other):
+        """Define inequiality behavior of request and response."""
+        return not self.__eq__(other)
+
     @property
     def text(self):
         return self.content
